#!/bin/bash

# --------------------------------------------------------------------------
# Retropak unit tests; part of retropak v0.1.
# Copyright (c) 2014 Ali Campbell

# This software is provided 'as-is', without any express or implied
# warranty. In no event will the authors be held liable for any damages
# arising from the use of this software.

# Permission is granted to anyone to use this software for any purpose,
# including commercial applications, and to alter it and redistribute it
# freely, subject to the following restrictions:

# 1. The origin of this software must not be misrepresented; you must not
#    claim that you wrote the original software. If you use this software
#    in a product, an acknowledgment in the product documentation would be
#    appreciated but is not required.
# 2. Altered source versions must be plainly marked as such, and must not be
#    misrepresented as being the original software.
# 3. This notice may not be removed or altered from any source distribution.
# --------------------------------------------------------------------------

# In order to run this test script, you will need the following tools:

# bash, mkdir, rm, rmdir, tee, wc, cmp, touch, dd, cp, ln, printf, sed, tr, cut, grep.

# On Windows this means installing MSYS.

BANNER="### Retropak unit tests. Part of retropak 0.1, (c) Ali Campbell 2014."

TESTPAK="test.pak"
TMPFILE1="tmp1"
TMPFILE2="tmp2"
TESTPAK_NOEXT="test_paks/no_extension"
TESTDIR="testdir"
TMPDIR1="foo"
TMPDIR2="bar"
RETROPAK="../retropak"
METAFILE="retropak.txt"
QUIT_ON_ERROR=1
UNPACK="unpack"
PACK="pack"
LOGFILE="log.txt"
MISSING_FILE="missing_file" # must not exist
TESTPAK_LEN_STANDARD_MP1=416
TESTPAK_STANDARD_MP1="test_paks/standard_mp1.pak"
TESTPAK_STANDARD_MP2="test_paks/standard_mp2.pak"
TESTPAK_EMPTY="test_paks/empty.pak"
TESTPAK_ZL_NAME="test_paks/zl_name.pak"
TESTPAK_ZL="test_paks/zl.pak"
TESTPAK_TRUNC="test_paks/trunc.pak"
TESTPAK_REVERSED_ENTRIES="test_paks/reversed_entries.pak"
TESTPAK_REVERSED_PAYLOADS="test_paks/reversed_payloads.pak"
TESTPAK_DUPLICATES="test_paks/duplicates.pak"
TESTMETA_STANDARD_MP1="test_metafiles/standard_mp1.txt"
TESTMETA_STANDARD_MP2="test_metafiles/standard_mp2.txt"
TESTMETA_REVERSED_ENTRIES="test_metafiles/reversed_entries.txt"
TESTMETA_REVERSED_PAYLOADS="test_metafiles/reversed_payloads.txt"
TESTMETA_EMPTY="test_metafiles/empty.txt"
TESTMETA_BAD_HEADER_FIELDLENS="test_metafiles/bad_header_fieldlens.txt"
TESTMETA_BAD_HEADER_LINE_LEN_LONG="test_metafiles/bad_header_line_len_long.txt"
TESTMETA_BAD_HEADER_LINE_LEN_SHORT="test_metafiles/bad_header_line_len_short.txt"
TESTMETA_BAD_HEADER_LINE_NULL="test_metafiles/bad_header_line_null.txt"
TESTMETA_BAD_HEADER_MAGIC="test_metafiles/bad_header_magic.txt"
TESTMETA_BAD_HEADER_PAK_TYPE="test_metafiles/bad_header_pak_type.txt"
TESTMETA_BAD_HEADER_PAK_VERSION="test_metafiles/bad_header_pak_version.txt"
TESTMETA_BAD_HEADER_RETROPAK_VERSION="test_metafiles/bad_header_retropak_version.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_1="test_metafiles/bad_entry_fieldlens_1.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_2="test_metafiles/bad_entry_fieldlens_2.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_3="test_metafiles/bad_entry_fieldlens_3.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_4="test_metafiles/bad_entry_fieldlens_4.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_5="test_metafiles/bad_entry_fieldlens_5.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_6="test_metafiles/bad_entry_fieldlens_6.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_7="test_metafiles/bad_entry_fieldlens_7.txt"
TESTMETA_BAD_ENTRY_FIELDLENS_8="test_metafiles/bad_entry_fieldlens_8.txt"
TESTMETA_PACK_LOWERCASE="test_metafiles/lowercase_pack.txt"
TESTMETA_PACK_SHORT_FILENAMES="test_metafiles/short_filenames_pack.txt"
TESTMETA_PACK_FILE_IS_SYMLINK="test_metafiles/pack_file_is_symlink.txt"

TESTPL_STANDARD_ZLIB_1="test_payloads/zlib_1"
TESTPL_STANDARD_ZLIB_2="test_payloads/zlib_2"
TESTPL_STANDARD_LZO_1="test_payloads/lzo_1"
TESTPL_STANDARD_LZO_2="test_payloads/lzo_2"
TESTPL_STANDARD_UNC_1="test_payloads/unc_1"
TESTPL_STANDARD_UNC_2="test_payloads/unc_2"
TESTPL_RAWS_DIR="test_payloads/raw"
TESTFILES_DIR="test_files"
TESTPAK_STANDARD_MP1_FILE1="aa11aa11_!File_1.TEST"
TESTPAK_STANDARD_MP1_FILE2="bb22bb22.L0VE"
TESTPAK_STANDARD_MP1_FILE3="cc33cc33.1234"
TESTPAK_STANDARD_MP1_FILE4="dd44dd44_!File_4.0XY9"
TESTPAK_STANDARD_MP2_FILE1="12345678.TEST"
TESTPAK_STANDARD_MP2_FILE2="87654321_!File_2.L0VE"
TESTPAK_STANDARD_MP2_FILE3="abbacafe_!File_3.1234"
TESTPAK_STANDARD_MP2_FILE4="deefaced.0XY9"

# hardcoded locations of things in the test PAK files
HACK_STANDARD_MP1_OFFSET_NUM_ENTRIES=$[  0x32 ]
HACK_STANDARD_MP1_OFFSET_NAME_LEN=$[     0x14 ]
HACK_STANDARD_MP1_OFFSET_NAME_START=$[   0x18 ]
HACK_STANDARD_MP1_OFFSET_NAME_MAGIC=$[   0x0c ]
HACK_STANDARD_MP1_OFFSET_NAME_ID=$[      0x10 ]
HACK_STANDARD_MP1_OFFSET_ZLIB_PAYLOAD=$[ 0xf7 ] # points somewhere into a zlib payload
HACK_STANDARD_MP1_OFFSET_ZLIB_UNC_LEN=$[ 0xe0 ] # points to decompressed block len in PAK
HACK_STANDARD_MP1_OFFSET_ENTRY_MAGIC=$[  0x3a ] # points to a magic value in the entries table
HACK_STANDARD_MP1_OFFSET_ENTRY_OFFSET=$[ 0x46 ] # points to an offset value in entries table
HACK_STANDARD_MP1_OFFSET_ENTRY_LEN=$[    0x42 ] # points to an offset value in entries table
HACK_STANDARD_MP2_OFFSET_LZO_PAYLOAD=$[ 0x12c ] # points somewhere into an LZO payload
HACK_STANDARD_MP2_OFFSET_LZO_UNC_LEN=$[ 0x120 ] # points to decompressed block len in PAK
HACK_STANDARD_MP2_OFFSET_LZO_SEGLEN=$[  0x124 ] # points to a LZO segsize in the PAK
HACK_DUPLICATES_OFFSET_PAYLOAD_PADDING=$[0x170] # points to padding after third instance of payload

PAYLOAD_FILE_ZLIB_1="test_payload_zlib_1"
PAYLOAD_FILE_ZLIB_2="test_payload_zlib_2"
PAYLOAD_FILE_UNC_1="test_payload_unc_1"
PAYLOAD_FILE_UNC_2="test_payload_unc_2"
PAYLOAD_FILE_LZO_1="test_payload_lzo_1"
PAYLOAD_FILE_LZO_2="test_payload_lzo_2"

PACK_INPUT_FILE_SYMLINK="test_files/eeee1111_for_testing_pack_file_is_symlink.SLNK"
PACK_INPUT_FILE_REAL="pack_file_is_symlink"

TESTPAK_STANDARD_MP1_FILE="${TESTPAK_STANDARD_MP1_FILE2}"

echo $BANNER
echo --- Retropak output is logged to log.txt, in the current directory.

MAC_OS_X=0
WINDOZE=0
DDFLAGS=""

# use bundled LZO library (Mac OS X):
if [ ! -z "`echo -n ${OSTYPE} | grep darwin`" ] ; then
  MAC_OS_X=1
  echo --- Mac OS X detected. Assuming bundled LZO in parent directory.
  export DYLD_LIBRARY_PATH=..
fi

if [ "${OS}" = "Windows_NT" -o "${OSTYPE}" = "msys" ] ; then
  WINDOZE=1
  echo --- Win32/MSYS detected.
  DDFLAGS="iflag=binary oflag=binary"
fi

function do_tests() {
  
  func_clean
  
  if [ ! -f "${RETROPAK}" ] ; then
    echo "Could not find retropak executable: ${RETROPAK}"
    exit 1
  fi
  
  # UNPACK: COMMAND-LINE RELATED
  # modes
  test_bad_modes                       # not pack/unpack
  # -p
  test_u_bad_p                         # unpack -p NULL, -p "" etc.
  test_u_non_existent_pakfile          # unpack -p does_not_exist
  test_u_pakfile_slashes_exist         # unpack -p ./foo/../bar/.///..//./valid.pak (kinda)
  test_u_pakfile_slashes_noexist       # unpack -p ./foo/../bar/.///..//./valid.pak (kinda)
  test_u_pakfile_is_dir                # unpack -p this_is_a_directory_actually
  test_u_pakfile_is_symlink            # unpack -p this_is_a_symlink_actually
  test_u_no_dirname_no_pak_extension   # unpack -p no_extension (fails, cannot work out dir)
  test_u_long_p                        # unpack --pakfile valid.pak
  # -d
  test_u_bad_d                         # unpack -p valid.pak -d ""
  test_u_have_dirname_no_pak_extension # unpack -p pak_no_extension -d valid_dir_name (succeeds)
  test_u_dir_slashes_exist             # unpack -p valid.pak -d ./foo/../bar/.///..//files
                                       #   (where foo, bar, files exist)
  test_u_dir_slashes_noexist           # unpack -p valid.pak -d ./foo/../bar/.///..//files
                                       #   (where foo, bar, files don't exist)
  test_u_long_d                        # --dirname
  test_u_dir_is_file                   # unpack -p valid.pak -d this_is_actually_a_file
  test_u_d_trailing_slash              # unpack -p valid.pak -d valid_dir_name/
  test_u_dir_is_symlink
  # -m
  test_u_bad_m
  test_u_working_m                     # needed because unlike -p and -d we don't use -m
                                       # in other tests
  # -o
  test_u_metafile_o                    # check metafile overwrite with and without -o
  test_u_o_metafile_is_symlink         # unpack -o -m some_symlink
  test_u_files_o                       # check unpacked files/dirs overwrite with and without -o
  test_u_metafile_is_dir_o             # unpack -o -p valid.pak -d testdir -m actually_a_dir
  # -v
  test_u_v                             # strlen(verbose output) > strlen(normal output)
  test_u_v_long                        # --verbose
  # -s
  test_u_s                             #
  test_u_s_long                        # --short-filenames
  # -l
  test_u_l                             #
  test_u_l_long                        # --lower-case-filenames
  # -c
  test_u_c                             # make PAK w/3 copies of same file and corrupt the final one
  test_u_c_long                        # --crc-duplicates
  # -r
  test_u_r
  test_u_r_long                        # --raw-payload
  # -x
  test_u_x
  test_u_x_long                        # --no-extract

  # UNPACK: PAK INTEGRITY
  # general
  test_u_zl_pak             # pakfile size = 0
  test_u_trunc              # only test truncation once because it's all fread() fail
  # top-related
  test_u_bad_version        # 0x00030005 fail
  test_u_bad_2nd_word       # 0x00000000 fail
  test_u_empty_pak          # simplest valid PAK (0x00030005 0x00000000 0x00000000 0x00000000)
  test_u_insane_num_names   # try number off end of file & negative number for num_names
  test_u_insane_num_entries # try number off end of file & negative number for num_entries
  test_u_insane_name_len    # try namelen=0x80000000
  test_u_zl_name            # test with name=""
  test_u_illegal_name       # a name table entry contains bad characters
  test_u_illegal_name_magic # a name table entry has bad magic (nulls?)
  test_u_lowercase_name_magic
  test_u_name_id_not_in_entries_table  # name table entry has non-included entry ID
  test_u_illegal_entry_magic
  test_u_lowercase_entry_magic
  test_u_bad_source_offsets   # zero, within-header, insane, misaligned offsets
  test_u_bad_source_lengths   # zero, insane, misaligned lengths
  # payload-related
  test_u_corrupt_zlib         # corrupt zlib payload
  test_u_bad_zlib_len         # payload length in PAK file too long, too short, totally insane (zlib)
  test_u_corrupt_lzo          # corrupt lzo segment (COULD CRASH?)
  test_u_bad_lzo_len          # payload length in PAK file too long, too short, totally insane (LZO)
  test_u_lzo_zero_segsize     # evil segment size for lzo segment
  #test_u_lzo_then_zlib       # first entry LZO, second entry zlib, should fail
  #test_u_zlib_then_lzo       # first entry zlib, second entry LZO, should fail
  
  # UNPACK: STANDARD FEATURES
  test_u_dir_creation           # check -d creates dir
  test_u_metafile_integrity_mp1 # check correct metafile is produced for standard PAKs
  test_u_metafile_integrity_mp2
  test_u_output_integrity_mp1   # ensure created files from standard paks are okay
  test_u_output_integrity_mp2
  test_u_reversed_payloads      # pak w/same entry table order, reversed payload order
  test_u_reversed_entries       # pak w/reversed entry table order
  
  # PACKING
  test_p_mp1
  test_p_mp2
  test_p_empty_pak           # metafile has header line only, this used to be a bug
  test_p_metafile_zl
  test_p_metafile_bad_header # multitest
  test_p_metafile_bad_entry  # multitest
  
  test_p_d_dir_is_file
  test_p_d_dir_is_symlink
  test_p_d_trailing_slash
  test_p_o                   # test overwriting output PAK
  test_p_o_pak_is_dir        # test failure of overwriting output PAK when exists and is dir
  test_p_o_pak_is_symlink    # test failure of overwriting output PAK when exists and is symlink
  test_p_l                   # lowercase filenames, should be straightforward enough
  test_p_s                   # short filenames, look at test_u_s
  test_p_dir_creation        # check -p creates containing path
  test_p_metafile_is_symlink # this is allowed
  test_p_file_is_symlink     # do we allow this or not? could make a case for either
  
  test_p_metafile_is_dir     # we definitely don't allow this.
  test_p_file_is_dir         # or this.
  
  test_p_derive_pak_and_meta_from_dirname  # discovered in 0.1 RC2
  
  echo "--- ALL TESTS COMPLETED."
  
}


function test_bad_modes() {
  # 1. retropak (FAIL)
  # 2. retropak "" (FAIL)
  # 3. retropak BAD (FAIL)
  func_retropak >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 1
  func_retropak "" >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 2
  func_retropak "BAD" >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 3
}

function test_u_bad_p() {
  # 1. retropak unpack (FAIL)
  # 2. retropak unpack -p (FAIL)
  # 3. retropak unpack -p "" (FAIL)
  func_retropak ${UNPACK} >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 1
  func_retropak ${UNPACK} -p >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 2
  func_retropak ${UNPACK} -p "" >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 3
}

function test_u_non_existent_pakfile() {
  # 1. retropak unpack -p missing_pakfile
  # FAIL
  func_retropak ${UNPACK} -p "${MISSING_FILE}" >> "${LOGFILE}"
  result_check 20 "${FUNCNAME}" 1
}

function test_u_pakfile_slashes_exist() {
  # 1. retropak unpack -p ./foo/../bar/.///..//./valid.pak
  # OK because foo and bar exist (func_mk_tmpdirs)
  func_rm_testdir
  func_mk_tmpdirs
  func_retropak \
    ${UNPACK} \
    -p "./${TMPDIR1}/../${TMPDIR2}/.///..//./${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_tmpdirs
  func_rm_testdir
}

function test_u_pakfile_slashes_noexist() {
  # 1. unpack -p ./foo/../bar/.///..//./valid.pak
  # FAIL because foo and bar don't exist
  func_rm_testdir
  func_rm_tmpdirs
  func_retropak \
    ${UNPACK} \
    -p "./${TMPDIR1}/../${TMPDIR2}/.///..//./${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  X=$? # yuck
  if [ "${WINDOZE}" = 1 ] ; then
    [ $X = 0 ]
    result_check 0 "${FUNCNAME}" 1 # windows is OK with this
  else
    [ $X = 6 ]
    result_check 0 "${FUNCNAME}" 1 # UNIX isn't
  fi
  func_rm_testdir
  func_rm_tmpdirs
}

function test_p_l() {
  func_rm_testpak
  # results depend on whether filesystem is case sensitive or not
  # currently this test assumes a case sensitive filesystem
  # try first without -l and then with -l
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_PACK_LOWERCASE}" >> "${LOGFILE}"
  X=$?
  if [ '(' "${MAC_OS_X}" == 1 ')' -o '(' "${WINDOZE}" = 1 ')' ] ; then
    [ $X = 0 ]
    result_check 0 "${FUNCNAME}" 1
  else
    [ $X = 46 ]
    result_check 0 "${FUNCNAME}" 1
  fi
  func_rm_testpak
  func_retropak ${PACK} \
                -l \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_PACK_LOWERCASE}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_testpak
}

function test_p_dir_creation() {
  func_rm_testpak
  func_rm_tmpdirs
  func_retropak ${PACK} \
                -p "${TMPDIR1}/${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TMPDIR1}/${TESTPAK}" "${TESTPAK_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_tmpdirs
  func_rm_testpak
}

function test_p_s() {
  func_rm_testpak
  # try first without -s and then with -s
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_PACK_SHORT_FILENAMES}" >> "${LOGFILE}"
  result_check 46 "${FUNCNAME}" 1
  func_rm_testpak
  func_retropak ${PACK} \
                -s \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_PACK_SHORT_FILENAMES}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_testpak
}


function test_p_o() {

  # 1. check it fails to overwrite PAK without -o
  # 2. check it succeeds with -o
  
  func_rm_testpak
  if ! touch "${TESTPAK}" ; then
    echo "test_p_o(): touch failed"
    exit 1
  fi
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 7 "${FUNCNAME}" 1
  
  func_retropak ${PACK} \
                -o \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 2
    
  func_rm_testpak
}

function test_p_file_is_symlink() {
  func_rm_testpak
  if [ -L "${PACK_INPUT_FILE_SYMLINK}" ] ; then
    rm "${PACK_INPUT_FILE_SYMLINK}" # FIXME
  fi
  if ! ln -s "${PACK_INPUT_FILE_REAL}" "${PACK_INPUT_FILE_SYMLINK}" ; then
    echo "test_p_file_is_symlink(): ln failed"
    exit 1
  fi
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_PACK_FILE_IS_SYMLINK}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testpak
  rm "${PACK_INPUT_FILE_SYMLINK}"
}


function test_p_file_is_dir() {
  func_rm_testpak
  # this file is referenced in the metafile used:
  if [ -d "${PACK_INPUT_FILE_SYMLINK}" ] ; then
    rmdir "${PACK_INPUT_FILE_SYMLINK}" # FIXME
  fi
  mkdir "${PACK_INPUT_FILE_SYMLINK}"
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_PACK_FILE_IS_SYMLINK}" >> "${LOGFILE}"
  result_check 36 "${FUNCNAME}" 1
  rmdir "${PACK_INPUT_FILE_SYMLINK}"
  func_rm_testpak
}


function test_p_metafile_is_symlink() {
  func_rm_testpak
  if [ -e "${TMPFILE1}" ] ; then
    rm "${TMPFILE1}"
  fi
  if ! ln -s "${TESTMETA_STANDARD_MP1}" "${TMPFILE1}" ; then
    echo "test_p_metafile_is_symlink(): ln failed"
    exit 1
  fi
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TMPFILE1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  rm "${TMPFILE1}"
  cmp "${TESTPAK}" "${TESTPAK_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_testpak
}

function test_p_metafile_is_dir() {
  func_rm_testpak
  func_rm_tmpdirs
  func_mk_tmpdirs

  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TMPDIR1}" >> "${LOGFILE}"
  result_check 36 "${FUNCNAME}" 1

  func_rm_tmpdirs
  
  func_rm_testpak
}

function test_p_o_pak_is_dir() {
  func_rm_testpak
  func_rm_tmpdirs
  func_mk_tmpdirs
  
  func_retropak ${PACK} \
                -o \
                -p "${TMPDIR1}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 6 "${FUNCNAME}" 1
  
  func_rm_tmpdirs
  func_rm_testpak
}

function test_p_o_pak_is_symlink() {

  if [ $WINDOZE = 1 ] ; then
    echo "SKIP (WIN32): ${FUNCNAME}"
    return
  fi

  # current behaviour: allow the output PAK file
  # to actually be a symlink, then when we open
  # it with -o we overwrite the linked file
  # (rather than replacing the symlink with a
  # file).

  func_rm_testpak
  func_rm_testdir
  
  if ! touch "${TESTPAK}" ; then
    echo "${FUNCNAME}: touch failed"
    exit 1
  fi
  
  if ! ln -s "${TESTPAK}" "${TESTDIR}" ; then
    echo "${FUNCNAME}: ln failed"
    exit 1
  fi
  
  func_retropak ${PACK} \
                -o \
                -p "${TESTDIR}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  
  rm "${TESTDIR}" # FIXME

  cmp "${TESTPAK}" "${TESTPAK_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2

  func_rm_testpak
}

function test_p_d_dir_is_file() {
  # 1. retropak pack -p valid.pak -d actually_a_file
  func_rm_testpak
  func_rm_testdir
  #rm "${TESTDIR}"

  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTDIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  X=$?
  func_rm_testdir
  [ $X = 46 ]
  result_check 0 "${FUNCNAME}" 1
  func_rm_testpak
}

function test_p_d_dir_is_symlink() {

  if [ $WINDOZE = 1 ] ; then
    echo "SKIP (WIN32): ${FUNCNAME}"
    return
  fi

  # 1. retropak pack -p valid.pak -d actually_a_symlink
  func_rm_testpak
  func_rm_testdir

  if ! ln -s "${TESTFILES_DIR}" "${TESTDIR}" ; then
    echo "${FUNCNAME}: ln failed"
    exit 1
  fi
  
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTDIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  X=$?
  rm "${TESTDIR}"
  [ $X = 0 ]
  result_check 0 "${FUNCNAME}" 1
  
  func_rm_testpak
  func_rm_testdir
}

function test_p_d_trailing_slash() {
  # 1. retropak pack -p valid.pak -d actually_a_symlink
  func_rm_testpak
  
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}/" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
 
  func_rm_testpak
}

function test_u_dir_is_file() {
  # 1. retropak unpack -p valid.pak -d actually_a_file
  # OK: foo, bar, files exist
  func_rm_testpak
  if ! touch "${TESTPAK}" ; then
    echo "test_u_dir_is_file(): touch failed"
    exit 1
  fi
  func_retropak \
    ${UNPACK} \
    -d "${TESTPAK}" \
    -p "${TESTPAK_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 22 "${FUNCNAME}" 1
  func_rm_testpak
}

function test_u_dir_creation() {
  # retropak unpack -p valid.pak -d tmpdir1/tmpdir2
  # OK
  func_rm_tmpdirs
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TMPDIR1}/${TMPDIR2}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -d "${TMPDIR1}/${TMPDIR2}" ] # check it worked
  result_check 0 "${FUNCNAME}" 2
  rm -rf "${TMPDIR1}" # FIXME: don't use rm -rf
  func_rm_testdir
}



function test_u_dir_slashes_exist() {
  # retropak unpack -p ./foo/../bar/.///..//./files
  # OK: foo, bar, files exist
  func_rm_testdir
  func_mk_tmpdirs
  func_retropak \
    ${UNPACK} \
    -p "./${TMPDIR1}/../${TMPDIR2}/.///..//./${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -e "${TESTDIR}" ] # check it worked
  result_check 0 "${FUNCNAME}" 2
  func_rm_tmpdirs
  func_rm_testdir
}

function test_u_dir_slashes_noexist() {
  # 1. unpack -p ./foo/../bar/.///..//./files
  # FAIL because foo and bar don't exist
  func_rm_testdir
  func_rm_tmpdirs
  func_retropak \
    ${UNPACK} \
    -p "./${TMPDIR1}/../${TMPDIR2}/.///..//./${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  X=$?
  if [ "${WINDOZE}" = 1 ] ; then
    [ $X = 0 ]
    result_check 0 "${FUNCNAME}" 1 # windows is OK with this
  else
    [ $X = 6 ]
    result_check 0 "${FUNCNAME}" 1 # unix isn't
  fi
  func_rm_testdir
  func_rm_tmpdirs
}

function test_u_long_p() {
  # 1. retropak unpack --pakfile valid.pak -d testdir
  # OK
  func_rm_testdir
  func_mk_tmpdirs
  func_retropak \
    ${UNPACK} \
    --pakfile "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -e "${TESTDIR}" ] # check it worked
  result_check 0 "${FUNCNAME}" 2
  func_rm_tmpdirs
  func_rm_testdir
}

function test_u_metafile_integrity_mp1() {
  # do a standard unpack, cmp metafile against provided copy
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp -b "${TESTDIR}"/"${METAFILE}" "${TESTMETA_STANDARD_MP1}" > /dev/null
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_metafile_integrity_mp2() {
  # do a standard unpack, cmp metafile against provided copy
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP2}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp -b "${TESTDIR}"/"${METAFILE}" "${TESTMETA_STANDARD_MP2}" > /dev/null
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_c() {
  # test CRCing for duplicate payloads
  func_rm_testdir
  if ! cp "${TESTPAK_DUPLICATES}" "${TESTPAK}" ; then
    echo "test_u_c(): cp failed"
    exit 1
  fi
  # modify the padding instead of the main payload,
  # so it doesn't mess up the compression
  func_edit_file "${TESTPAK}" \
                 ${HACK_DUPLICATES_OFFSET_PAYLOAD_PADDING} \
                 $'\xab' $'\xcd' $'\xef' $'\x01'
  # first try it without -c and check it works:
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
  # then try it with -c and check it fails:
  func_retropak \
    ${UNPACK} \
    -c \
    -p "${TESTPAK}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 14 "${FUNCNAME}" 2
  func_rm_testpak
  func_rm_testdir
}

function test_u_c_long() {
  # test CRCing for duplicate payloads, long mode
  func_rm_testdir
  if ! cp "${TESTPAK_DUPLICATES}" "${TESTPAK}" ; then
    echo "test_u_c(): cp failed"
    exit 1
  fi
  # modify the padding instead of the main payload,
  # so it doesn't mess up the compression
  func_edit_file "${TESTPAK}" \
                 ${HACK_DUPLICATES_OFFSET_PAYLOAD_PADDING} \
                 $'\xab' $'\xcd' $'\xef' $'\x01'
  # first try it without -c and check it works:
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
  # then try it with -c and check it fails:
  func_retropak \
    ${UNPACK} \
    --crc-duplicates \
    -p "${TESTPAK}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 14 "${FUNCNAME}" 2
  func_rm_testpak
  func_rm_testdir
}

function test_u_l() {
  # lower case filenames
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -l \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  LC=`echo "${TESTPAK_STANDARD_MP1_FILE1}" | tr 'A-Z' 'a-z'`
  [ -f "${TESTDIR}"/"${LC}" ]
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_l_long() {
  # lower case filenames
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    --lower-case-filenames \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  LC=`echo "${TESTPAK_STANDARD_MP1_FILE1}" | tr 'A-Z' 'a-z'`
  [ -f "${TESTDIR}"/"${LC}" ]
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_s() {
  # short filenames
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -s \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE1}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE1}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 1
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE2}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE2}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 2
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE3}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE3}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 3
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE4}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE4}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 4
  func_rm_testdir
}

function test_u_s_long() {
  # short filenames
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    --short-filenames \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE1}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE1}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 1
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE2}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE2}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 2
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE3}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE3}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 3
  NAME=`echo "${TESTPAK_STANDARD_MP1_FILE4}" | grep -oe '^[^_.]\+'`
  EXT=`echo "${TESTPAK_STANDARD_MP1_FILE4}" | grep -oe '[^.]\{4\}$'`
  [ -f "${TESTDIR}"/"${NAME}.${EXT}" ]
  result_check 0 "${FUNCNAME}" 4
  func_rm_testdir
}

function test_u_output_integrity_mp1() {
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  # compare file 1
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" > /dev/null
  result_check 0 "${FUNCNAME}" 2
  # compare file 2
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE2}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP1_FILE2}" > /dev/null
  result_check 0 "${FUNCNAME}" 3
  # compare file 3
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE3}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP1_FILE3}" > /dev/null
  result_check 0 "${FUNCNAME}" 4
  # compare file 4
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE4}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP1_FILE4}" > /dev/null
  result_check 0 "${FUNCNAME}" 5
  func_rm_testdir
}

function test_u_output_integrity_mp2() {
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP2}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  # compare file 1
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP2_FILE1}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP2_FILE1}" > /dev/null
  result_check 0 "${FUNCNAME}" 2
  # compare file 2
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP2_FILE2}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP2_FILE2}" > /dev/null
  result_check 0 "${FUNCNAME}" 3
  # compare file 3
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP2_FILE3}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP2_FILE3}" > /dev/null
  result_check 0 "${FUNCNAME}" 4
  # compare file 4
  cmp -b "${TESTDIR}"/"${TESTPAK_STANDARD_MP2_FILE4}" \
         "${TESTFILES_DIR}"/"${TESTPAK_STANDARD_MP2_FILE4}" > /dev/null
  result_check 0 "${FUNCNAME}" 5
  func_rm_testdir
}

function test_u_v() {
  # 1. retropak unpack -p valid.pak --directory testdir
  # OK
  func_rm_testdir
  LEN_NORMAL=`func_retropak \
              ${UNPACK} \
              -p "${TESTPAK_STANDARD_MP1}" \
              -d "${TESTDIR}" | tee -a "${LOGFILE}" | wc -c`
  func_rm_testdir
  LEN_VERBOSE=`func_retropak \
              ${UNPACK} \
              -v \
              -p "${TESTPAK_STANDARD_MP1}" \
              -d "${TESTDIR}" | tee -a "${LOGFILE}" | wc -c`
  [ ${LEN_NORMAL} -lt ${LEN_VERBOSE} ] # compare lengths
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
}

function test_u_v_long() {
  # 1. retropak unpack -p valid.pak --directory testdir
  # OK
  func_rm_testdir
  LEN_NORMAL=`func_retropak \
              ${UNPACK} \
              -p "${TESTPAK_STANDARD_MP1}" \
              -d "${TESTDIR}" | tee -a "${LOGFILE}" | wc -c`
  func_rm_testdir
  LEN_VERBOSE=`func_retropak \
              ${UNPACK} \
              --verbose \
              -p "${TESTPAK_STANDARD_MP1}" \
              -d "${TESTDIR}" | tee -a "${LOGFILE}" | wc -c`   
  [ ${LEN_NORMAL} -lt ${LEN_VERBOSE} ] # compare lengths
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
}

function test_u_long_d() {
  # 1. retropak unpack -p valid.pak --directory testdir
  # OK
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    --directory "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -d "${TESTDIR}" ] # check it worked
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_dir_is_symlink() {

  if [ "${WINDOZE}" = 1 ] ; then
    echo "SKIP (WIN32): ${FUNCNAME}"
    return
  fi

  func_rm_testdir
  func_rm_tmpdirs
  func_mk_tmpdirs
  if ! ln -s "${TMPDIR1}" "${TESTDIR}" ; then
    echo "test_u_dir_is_symlink(): ln failed"
    exit 1
  fi
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -f "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE1}" ] # check it worked
  result_check 0 "${FUNCNAME}" 2
  if ! rm "${TESTDIR}" ; then
    echo "test_u_dir_is_symlink(): rm failed"
    exit 1
  fi
  func_rm_tmpdirs
}

function test_u_d_trailing_slash() {
  # 1. retropak unpack -p valid.pak -d testdir/
  # OK
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    --directory "${TESTDIR}/" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -d "${TESTDIR}" ] # check it worked
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_pakfile_is_dir() {
  func_rm_testdir
  func_mk_tmpdirs
  #if ! mkdir ${TESTDIR} ; then  # we'll borrow this
  #  echo "test_u_pakfile_is_dir(): mkdir failed"
  #  exit 1
  #fi
  func_retropak ${UNPACK} -p "${TMPDIR1}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 70 "${FUNCNAME}" 1
  func_rm_tmpdirs
  func_rm_testdir
}

function test_u_pakfile_is_symlink() {
  # this is okay, we don't mind reading from symlinks
  func_rm_testpak
  func_rm_testdir
  if ! ln -s "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then  # we'll borrow this
    echo "test_u_pakfile_is_symlink(): ln failed"
    exit 1
  fi
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testpak
  func_rm_testdir
}

function test_u_r() {
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -r \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE1}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE1}"
  result_check 0 "${FUNCNAME}" 2
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE2}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE2}"
  result_check 0 "${FUNCNAME}" 3
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE3}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE3}"
  result_check 0 "${FUNCNAME}" 4
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE4}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE4}"
  result_check 0 "${FUNCNAME}" 5
  func_rm_testdir
}

function test_u_r_long() {
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    --raw-payload \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE1}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE1}"
  result_check 0 "${FUNCNAME}" 2
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE2}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE2}"
  result_check 0 "${FUNCNAME}" 3
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE3}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE3}"
  result_check 0 "${FUNCNAME}" 4
  cmp "${TESTDIR}/${TESTPAK_STANDARD_MP1_FILE4}" \
      "${TESTPL_RAWS_DIR}"/"${TESTPAK_STANDARD_MP1_FILE4}"
  result_check 0 "${FUNCNAME}" 5
  func_rm_testdir
}

function test_u_metafile_o() {
  # test metafile overwrite
  func_rm_testdir
  touch "${METAFILE}"
  func_retropak \
    ${UNPACK} \
    -o \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" \
    -m "${METAFILE}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -s "${METAFILE}" ]
  result_check 0 "${FUNCNAME}" 2
  if [ -e "${METAFILE}" ] ; then rm "${METAFILE}" ; fi # FIXME
  func_rm_testdir
}

function test_u_o_metafile_is_symlink() {

  if [ "${WINDOZE}" = 1 ] ; then
    echo "SKIP (WIN32): ${FUNCNAME}"
    return
  fi

  # we don't want it trying to write to symlinks ...
  func_rm_testdir
  if [ -L "${TMPFILE1}" ] ; then
    if ! rm "${TMPFILE1}" ; then
      echo "test_u_o_metafile_is_symlink(): rm [1] failed"
      exit 1
    fi
  fi
  touch "${METAFILE}"
  if ! ln -s "${METAFILE}" "${TMPFILE1}" ; then
    echo "test_u_o_metafile_is_symlink(): ln failed"
    exit 1
  fi
  func_retropak \
    ${UNPACK} \
    -o \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" \
    -m "${TMPFILE1}" >> "${LOGFILE}"
  result_check 36 "${FUNCNAME}" 1
  if ! rm "${TMPFILE1}" ; then
    echo "test_u_o_metafile_is_symlink(): rm [2] failed"
    exit 1
  fi
  if [ -e "${METAFILE}" ] ; then rm "${METAFILE}" ; fi # FIXME
  func_rm_testdir
}

function test_u_files_o() {
  # check unpacked files overwrite with and without -o
  func_rm_testdir
  if ! mkdir "${TESTDIR}" ; then echo "test_u_files_o(): mkdir failed" ; exit 1 ; fi
  touch "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE}"
  func_retropak ${UNPACK} -p "${TESTPAK_STANDARD_MP1}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 37 "${FUNCNAME}" 1
  func_retropak ${UNPACK} -o -p "${TESTPAK_STANDARD_MP1}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 2
  # check it got bigger
  [ -s "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE}" ]
  result_check 0 "${FUNCNAME}" 3
  func_rm_testdir
}

function test_u_metafile_is_dir_o() {
  # try to overwrite a directory with the metafile
  func_rm_testdir
  func_mk_tmpdirs
  func_retropak \
    ${UNPACK} \
    -o \
    -p "${TESTPAK_STANDARD_MP1}" \
    -m "${TMPDIR1}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 36 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_tmpdirs
}

function test_u_no_dirname_no_pak_extension() {
  # 1. retropak unpack -p noextension
  # FAIL (because dirname cannot be computed)
  func_retropak ${UNPACK} -p "${TESTPAK_NOEXT}" >> "${LOGFILE}"
  result_check 20 "${FUNCNAME}" 1
}

function test_u_have_dirname_no_pak_extension() {
  # the successful version of test_u_no_dirname_no_pak_extension
  # 1. retropak unpack -p <name> where <name> has no extension but -d *is* provided
  # OK
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_NOEXT}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
}

function test_u_bad_m() {
  # 1. retropak unpack -p valid.pak -d testdir -m
  # 2. retropak unpack -p valid.pak -d testdir -m ""
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" \
    -m >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 1
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" \
    -m "" >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 2
}

function test_u_working_m() {
  func_rm_testdir
  func_mk_tmpdirs
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_STANDARD_MP1}" \
    -d "${TESTDIR}" \
    -m "${TMPDIR1}//${METAFILE}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -f "${TMPDIR1}/${METAFILE}" ]
  result_check 0 "${FUNCNAME}" 2
  func_rm_tmpdirs
  func_rm_testdir
}

function test_u_empty_pak() {
  # 1. retropak unpack -p testpak_empty.pak (OK)
  # 2. expect metafile created
  # 3. expect no output directory created
  META="${TESTDIR}"/"${METAFILE}"
  check_metafile "${META}" # make sure it's deleted
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_EMPTY}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  # simplest case ought still to produce a metafile.
  # currently it doesn't, so this test fails
  check_metafile "${METAFILE}" # deletes it
  echo "--- FIXME: TEST SKIPPED because retropak should create metafile for empty PAK, doesn't:"
  echo SKIP: "${FUNCNAME}"_2
  #result_check 0 "${FUNCNAME}" 2
  # should the output directory be created if no files
  # are to be extracted?
  # for now let's say no
  [ ! -d "${OPDIR}" ]
  result_check 0 "${FUNCNAME}" 3
  # clean up
  func_rm_testdir
}

function test_u_zl_pak() {
  # 1. retropak unpack -p testpak_zl.pak (FAIL)
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_ZL}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 13 "${FUNCNAME}" 1
  func_rm_testdir
}

function test_u_trunc() {
  # general truncation test, code 13
  # (we only do this once because otherwise it would get stupid,
  #  and it's always code 13 anyway)
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_TRUNC}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 13 "${FUNCNAME}" 1
  #X=`echo "${TESTPAK_TRUNC}" | func_dir_from_pakname`
  #if [ -e "${X}" ] ; then rm "${X}"/* 2> /dev/null ; rmdir "${X}" ; fi
  func_rm_testdir
}

function test_u_d() {
  # make sure -d works and dir is created
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_STANDARD_MP1}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  [ -d "${TESTDIR}" ]
  result_check 0 "${FUNCNAME}" 2
  func_rm_testdir
}

function test_u_reversed_payloads() {
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_REVERSED_PAYLOADS}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  cmp -b "${TESTDIR}"/"${METAFILE}" "${TESTMETA_REVERSED_PAYLOADS}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
}

function test_u_reversed_entries() {
  func_rm_testdir
  func_retropak \
    ${UNPACK} \
    -p "${TESTPAK_REVERSED_ENTRIES}" \
    -d "${TESTDIR}" >> "${LOGFILE}"
  cmp -b "${TESTDIR}"/"${METAFILE}" "${TESTMETA_REVERSED_ENTRIES}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
}

function test_u_bad_d() {
  # 1. retropak unpack -p test.pak -d
  # 2. retropak unpack -p test.pak -d ""
  func_retropak ${UNPACK} -p "${TESTPAK_STANDARD_MP1}" -d >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 1
  func_retropak ${UNPACK} -p "${TESTPAK_STANDARD_MP1}" -d "" >> "${LOGFILE}"
  result_check 2 "${FUNCNAME}" 2
  X=`echo "${TESTPAK_STANDARD_MP1}" | func_dir_from_pakname` # potential dir created
  if [ -e "${X}" ] ; then rm "${X}"/* 2> /dev/null ; rmdir "${X}" ; fi
}

function test_u_bad_version() {
  # 1. PAK has bad first word (0x00030005)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_bad_version(): cp failed"
    exit 1
  fi
  func_edit_file "${TESTPAK}" 0 $'\xff' $'\xff' $'\xff' $'\xff'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 54 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_bad_2nd_word() {
  # 1. PAK has bad second word (NULL)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_bad_2nd_word(): cp failed"
    exit 1
  fi
  func_edit_file "${TESTPAK}" 4 $'\xff' $'\xff' $'\xff' $'\xff'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 25 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_insane_num_names() {
  # 1. modify test PAK to have 0x80000000 names
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_insane_num_names(): cp failed"
    exit 1
  fi
  func_edit_file "${TESTPAK}" 8 $'\x80' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 26 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_insane_num_entries() {
  # 1. modify test PAK to have 0x80000000 entries
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_insane_num_entries(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NUM_ENTRIES} \
    $'\x80' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 27 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_insane_name_len() {
  # 1. modify test PAK to have a name of length 0x80000000
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_insane_name_len(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NAME_LEN} \
    $'\x80' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 29 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_zl_name() {
  # 1. modify test PAK to have a name of length 0
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_zl_name(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NAME_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 71 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_x() {
  # don't actually extract
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_STANDARD_MP1}" -d "${TESTDIR}" -x >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTDIR}"/"${METAFILE}" "${TESTMETA_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 3
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 4
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 5
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 6
  func_rm_testdir
}

function test_u_x_long() {
  # don't actually extract
  func_rm_testdir
  func_retropak ${UNPACK} -p "${TESTPAK_STANDARD_MP1}" -d "${TESTDIR}" --no-extract >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTDIR}"/"${METAFILE}" "${TESTMETA_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 3
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 4
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 5
  [ ! -e "${TESTDIR}"/"${TESTPAK_STANDARD_MP1_FILE1}" ]
  result_check 0 "${FUNCNAME}" 6
  func_rm_testdir
}

function test_u_illegal_name() {
  # 1. modify test PAK to have an illegal name
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_illegal_name(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NAME_START} \
    $'\x80' $'\x0d' $'\xff' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 72 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_illegal_name_magic() {
  # 1. magic *in name table* has illegal character
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_illegal_name_magic(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NAME_MAGIC} \
    $'\x80' $'\x0d' $'\xff' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 28 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_illegal_entry_magic() {
  # 1. magic in entry table has illegal character
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_illegal_entry_magic(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_MAGIC} \
    $'\x80' $'\x0d' $'\xff' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 75 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_lowercase_name_magic() {
  # 1. magic *in name table* is lower case
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_lowercase_name_magic(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NAME_MAGIC} \
    t e s t
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 28 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_lowercase_entry_magic() {
  # 1. magic *in name table* is lower case
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_lowercase_entry_magic(): cp failed"
    exit 1
  fi
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_MAGIC} \
    t e s t
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 75 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_bad_source_offsets() {
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_bad_source_offsets(): cp failed"
    exit 1
  fi
  # offset to payload = 0
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_OFFSET} \
    $'\x00' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 35 "${FUNCNAME}" 1
  func_rm_testdir
  # offset to payload = still inside header
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_OFFSET} \
    $'\x00' $'\x00' $'\x00' $'\x9c'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 35 "${FUNCNAME}" 2
  func_rm_testdir
  # offset to payload is misaligned
  # FIXME: this wouldn't unpack
  # correctly even if we did allow arbitrary alignments,
  # so it's kind of a cheat
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_OFFSET} \
    $'\x00' $'\x00' $'\x00' $'\xa1'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 35 "${FUNCNAME}" 3
  func_rm_testdir
  # offset to payload = beyond end of file
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_OFFSET} \
    $'\x00' $'\x00' $'\x01' $'\xa0'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 35 "${FUNCNAME}" 4
  func_rm_testdir
  # offset to payload = insane
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_OFFSET} \
    $'\x80' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 35 "${FUNCNAME}" 5
  func_rm_testdir
  func_rm_testpak
}

function test_u_bad_source_lengths() {
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_bad_source_lengths(): cp failed"
    exit 1
  fi
  # payload len = 0
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 34 "${FUNCNAME}" 1
  func_rm_testdir
  # payload len is misaligned
  # FIXME: this wouldn't unpack
  # correctly even if we did allow arbitrary alignments,
  # so it's kind of a cheat
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x3f'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 34 "${FUNCNAME}" 2
  func_rm_testdir
  # length = beyond end of file
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_LEN} \
    $'\x00' $'\x00' $'\x01' $'\x20' # FIXME: should be \x20 but func_edit_file is buggy
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 35 "${FUNCNAME}" 3
  func_rm_testdir
  # length = insane
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ENTRY_LEN} \
    $'\x80' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 34 "${FUNCNAME}" 4
  func_rm_testdir
  func_rm_testpak
}

function test_u_name_id_not_in_entries_table() {
  # 1. a name table entry has an ID that isn't in the entries table
  # (this is OK)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_name_id_not_in_entries_table(): cp failed"
    exit 1
  fi
  # edit PAK, rewrite name ID
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_NAME_ID} \
    $'\xbe' $'\xef' $'\xca' $'\xfe'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_corrupt_zlib() {
  # corrupt the zlib payload (MAY CRASH?)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_corrupt_zlib(): cp failed"
    exit 1
  fi
  # edit PAK, corrupt payload
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ZLIB_PAYLOAD} \
    $'\xbe' $'\xef' $'\xca' $'\xfe'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  # could be any of the zlib errors, 38-43
  X=$?
  [ '( ${X} -gt 38 ) -a ( ${X} -lt 43 )' ]
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_corrupt_lzo() {
  # corrupt the lzo payload (MAY CRASH?)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP2}" "${TESTPAK}" ; then
    echo "test_u_corrupt_lzo(): cp failed"
    exit 1
  fi
  # edit PAK, corrupt payload
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_PAYLOAD} \
    $'\xbe' $'\xef' $'\xca' $'\xfe'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  # could be any of the LZO errors, 59-65
  X=$?
  [ '( ${X} -gt 59 ) -a ( ${X} -lt 65 )' ]
  result_check 0 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_lzo_zero_segsize() {
  # corrupt the lzo segsize
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP2}" "${TESTPAK}" ; then
    echo "test_u_corrupt_lzo(): cp failed"
    exit 1
  fi
  # edit PAK, zero segsize
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_SEGLEN} \
    $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 74 "${FUNCNAME}" 1
  func_rm_testdir
  func_rm_testpak
}

function test_u_bad_zlib_len() {
  # uncompressed zlib block length in PAK file too long, too short, totally insane (zlib)
  # 1. too long (but within file length)
  # 2. too long (outside file length)
  # 3. zero
  # 4. too short (4)
  # 5. insane (0x80000000)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP1}" "${TESTPAK}" ; then
    echo "test_u_corrupt_zlib(): cp failed"
    exit 1
  fi
  # edit PAK, corrupt zlib decompressed len
  # 1
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ZLIB_UNC_LEN} \
    $'\x00' $'\x00' $'\x01' $'\x70'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 1
  func_rm_testdir
  # 2
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ZLIB_UNC_LEN} \
    $'\x00' $'\x00' $'\x01' $'\x90'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 2
  func_rm_testdir
  # 3
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ZLIB_UNC_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 3
  func_rm_testdir
  # 4
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ZLIB_UNC_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x04'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 4
  func_rm_testdir
  # 5
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP1_OFFSET_ZLIB_UNC_LEN} \
    $'\x80' $'\x00' $'\x00' $'\x04'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 5
  func_rm_testdir
  func_rm_testpak
}

function test_u_bad_lzo_len() {
  # uncompressed LZO block length in PAK file too long, too short, totally insane
  # 1. too long (but within file length)
  # 2. too long (outside file length)
  # 3. zero
  # 4. too short (4)
  # 5. insane (0x80000000)
  func_rm_testpak
  func_rm_testdir
  if ! cp "${TESTPAK_STANDARD_MP2}" "${TESTPAK}" ; then
    echo "test_u_corrupt_zlib(): cp failed"
    exit 1
  fi
  # edit PAK, corrupt zlib decompressed len
  # 1
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_UNC_LEN} \
    $'\x00' $'\x00' $'\x01' $'\x70'
#hexdump -C  "${TESTPAK}"
#exit 1
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 1
  func_rm_testdir
  # 2
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_UNC_LEN} \
    $'\x00' $'\x00' $'\x01' $'\x90'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 2
  func_rm_testdir
  # 3
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_UNC_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x00'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 3
  func_rm_testdir
  # 4
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_UNC_LEN} \
    $'\x00' $'\x00' $'\x00' $'\x04'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 4
  func_rm_testdir
  # 5
  func_edit_file \
    "${TESTPAK}" \
    ${HACK_STANDARD_MP2_OFFSET_LZO_UNC_LEN} \
    $'\x80' $'\x00' $'\x00' $'\x04'
  func_retropak ${UNPACK} -p "${TESTPAK}" -d "${TESTDIR}" >> "${LOGFILE}"
  result_check 73 "${FUNCNAME}" 5
  func_rm_testdir
  func_rm_testpak
}

function test_p_mp1() {
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTPAK}" "${TESTPAK_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_testpak
}


function test_p_derive_pak_and_meta_from_dirname() {
  func_mk_tmpdirs
  # copy all the test files into a new directory -- we'll use TMPDIR1
  if ! cp -R "${TESTFILES_DIR}"/* "${TMPDIR1}" ; then
    echo "${FUNCNAME}: cp [1] failed"
    func_rm_tmpdirs
    exit 1
  fi
  # also copy a metafile in there and give it the standard name
  if ! cp "${TESTMETA_STANDARD_MP1}" "${TMPDIR1}/retropak.txt" ; then
    echo "${FUNCNAME}: cp [2] failed"
    func_rm_tmpdirs
    exit 1
  fi
  # delete any existing ${TMPDIR1}.pak
  if [ -e "${TMPDIR1}.pak" ] ; then
    if ! rm "${TMPDIR1}.pak" ; then
      echo "${FUNCNAME}: rm failed"
      func_rm_tmpdirs
    fi
  fi
  func_retropak ${PACK} -d "${TMPDIR1}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  # check the generated PAK is as expected
  cmp "${TMPDIR1}.pak" "${TESTPAK_STANDARD_MP1}"
  result_check 0 "${FUNCNAME}" 2
  rm "${TMPDIR1}.pak"
  func_rm_tmpdirs
}



function test_p_mp2() {
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_STANDARD_MP2}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTPAK}" "${TESTPAK_STANDARD_MP2}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_testpak
}

function test_p_empty_pak() {
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_EMPTY}" >> "${LOGFILE}"
  result_check 0 "${FUNCNAME}" 1
  cmp "${TESTPAK}" "${TESTPAK_EMPTY}"
  result_check 0 "${FUNCNAME}" 2
  func_rm_testpak
}

function test_p_metafile_zl() {
  func_rm_testpak
  check_metafile "${METAFILE}"
  if ! touch "${METAFILE}" ; then echo "test_p_metafile_zl(): touch failed" ; exit 1 ; fi
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${METAFILE}" >> "${LOGFILE}"
  result_check 51 "${FUNCNAME}" 1
  check_metafile "${METAFILE}"
  func_rm_testpak
}

function test_p_metafile_bad_entry() {
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_1}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 1
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_2}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 2
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_3}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 3
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_4}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 4
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_5}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 5
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_6}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 6
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_7}" >> "${LOGFILE}"
  result_check 52 "${FUNCNAME}" 7
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_ENTRY_FIELDLENS_8}" >> "${LOGFILE}"
  # long filename might fail in a few ways:
  # - metadata line too long (48)
  # - file name too long (52)
  # - file name too long for OS, fopen() fails (47)
  # so for the time being it's enough to check that it fails
  [ $? != 0 ]
  result_check 0 "${FUNCNAME}" 8
  func_rm_testpak
}

function test_p_metafile_bad_header() {
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_FIELDLENS}" >> "${LOGFILE}"
  result_check 51 "${FUNCNAME}" 1
  #echo "SKIP: test_p_metafile_bad_header_1"
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_LINE_LEN_LONG}" >> "${LOGFILE}"
  result_check 49 "${FUNCNAME}" 2
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_LINE_LEN_SHORT}" >> "${LOGFILE}"
  result_check 49 "${FUNCNAME}" 3
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_LINE_NULL}" >> "${LOGFILE}"
  result_check 76 "${FUNCNAME}" 4
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_MAGIC}" >> "${LOGFILE}"
  result_check 51 "${FUNCNAME}" 5
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_PAK_TYPE}" >> "${LOGFILE}"
  result_check 57 "${FUNCNAME}" 6
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_PAK_VERSION}" >> "${LOGFILE}"
  result_check 54 "${FUNCNAME}" 7
  func_rm_testpak
  func_retropak ${PACK} \
                -p "${TESTPAK}" \
                -d "${TESTFILES_DIR}" \
                -m "${TESTMETA_BAD_HEADER_RETROPAK_VERSION}" >> "${LOGFILE}"
  result_check 55 "${FUNCNAME}" 8
  func_rm_testpak
}

# ========================================
# ========================================

function func_rm_testpak() {
  # testpak is used when we need to create an edited pak
  if [ -e "${TESTPAK}" ] ; then
    rm "${TESTPAK}"
  fi
}

function func_edit_file() {
  # corrupt some bytes of a file
  # <filename> <offset> <bytes ...>
  if [ -f "${TMPFILE1}" ] ; then
    if ! rm "${TMPFILE1}" ; then
      echo "func_edit_file(): could not delete ${TMPFILE1}"
      exit 1
    fi
  fi
  if [ -f "${TMPFILE2}" ] ; then
    if ! rm "${TMPFILE2}" ; then
      echo "func_edit_file(): could not delete ${TMPFILE2}"
      exit 1
    fi
  fi
  if ! mv "${1}" "${TMPFILE1}" ; then
    echo "func_edit_file(): could not mv ${1} to ${TMPFILE1}"
    exit 1
  fi
  OFF=${2}
  THEFILE="${1}"
  SEQLEN=$[ $# - 2 ] # num bytes
  if [ ${OFF} -ne 0 ] ; then
  #echo $DDFLAGS ; exit
    if ! dd ${DDFLAGS} if="${TMPFILE1}" of="${THEFILE}" bs=1 count=${OFF} 2> /dev/null  ; then
      echo "func_edit_file(): dd1 failed"
      exit 1
    fi
  fi
#exit 1
  shift 2
  X=0
  while [ ${X} -lt ${SEQLEN} ] ; do
    if [ -z "${1}" ] ; then
      #echo LOL NULL BYTE
      dd ${DDFLAGS} if=/dev/zero count=1 bs=1 >> "${THEFILE}" 2> /dev/null
    else
      #echo LOL REAL BYTE
      echo -n ${1} >> "${THEFILE}"
    fi 
    #  printf %c ${1} >> "${THEFILE}"
    shift
    X=$[ ${X} + 1 ]
  done
  OFF=$[ ${OFF} + ${SEQLEN} ]
  if [ ${OFF} -ne 0 ] ; then
    if ! dd ${DDFLAGS} if="${TMPFILE1}" of="${TMPFILE2}" bs=1 skip=${OFF} 2> /dev/null ; then
      echo "func_edit_file(): dd2 failed"
      exit 1
    fi
  fi
  cat "${TMPFILE2}" >> "${THEFILE}"
  if ! rm "${TMPFILE1}" ; then
    echo "func_edit_file(): rm ${TMPFILE1} failed."
    exit 1
  fi
  if ! rm "${TMPFILE2}" ; then
    echo "func_edit_file(): rm ${TMPFILE2} failed."
    exit 1
  fi
}

function func_rm_testdir() {
  func_rm_dir "${TESTDIR}"
}

function func_rm_dir() {
  if [ -e "${1}" ] ; then
    if [ ! -d "${1}" ] ; then
      echo "func_rm_dir(): ${1} was not a directory, aborting."
      exit 1
    fi
    rm "${1}"/* 2> /dev/null
    rmdir "${1}"
    if [ $? != 0 ] ; then
      echo "func_rm_dir(): ${1} could not be deleted, aborting."
      exit 1
    fi
  fi
}

function check_metafile() {
  if [ ! -f "${1}" ] ; then
    return 1
  else
    if ! rm "${1}" ; then
      echo "Could not delete metafile."
      exit 1
    fi
    return 0
  fi
}

function func_mkheader() {
  if [ -e ${1} ] ; then
    if ! rm "${1}" ; then
      echo "func_mkheader(): rm ${1} failed"
      exit 1
    fi
  fi
  printf %c%c%c%c $'\x00' $'\x03' $'\x00' $'\x05' > "${1}"
}

function func_mkpak_zl() {
  if [ -e ${1} ] ; then
    if ! rm "${1}" ; then
      echo "func_mkpak_zl(): rm ${1} failed"
      exit 1
    fi
  fi
  touch "${1}"
}

function func_mkpak_trunc_1() {
  if [ -e ${1} ] ; then
    if ! rm "${1}" ; then
      echo "func_mkpak_trunc_1(): rm ${1} failed"
      exit 1
    fi
  fi
  printf %c%c%c%c $'\x00' $'\x03' $'\x00' > "${1}"
}

#function func_mkpak_trunc_2() {
#  rm "${1}"
#  printf %c%c%c%c $'\x00' $'\x03' $'\x00' $'\x05' > "${1}"
#  printf %c%c%c%c $'\x00' $'\x00' $'\x00' >> "${1}"
#}

function func_mkpak_empty() {
  if [ -e ${1} ] ; then
    if ! rm "${1}" ; then
      echo "func_mkpak_empty(): rm ${1} failed"
      exit 1
    fi
  fi
  func_mkheader "${1}"
  printf %c%c%c%c $'\x00' $'\x00' $'\x00' $'\x00' >> "${1}" # no names
  printf %c%c%c%c $'\x00' $'\x00' $'\x00' $'\x00' >> "${1}" # no entries
  if [ ! -f "${1}" ] ; then
    echo "Empty test PAK could not be created."
    exit 1
  fi
}


function result_check() {
  X=$?
  if [ $1 != $X ] ; then
    echo "FAIL: $2_${3} [$X, wanted $1]"
    if [ $QUIT_ON_ERROR == 1 ] ; then exit 1 ; fi
  else
    echo "OK: ${2}_${3}"
  fi
}

function func_retropak() {
  echo "${RETROPAK}" "$@" >> "${LOGFILE}"
  "${RETROPAK}" "$@"
}

function func_mk_tmpdirs() {
  # FIXME: chex
  if [ ! -d "${TMPDIR1}" ] ; then
    if ! mkdir "${TMPDIR1}" ; then
      echo "func_mk_tmpdirs(): mkdir ${TMPDIR1} failed"
      exit 1
    fi
  fi
  if [ ! -d "${TMPDIR2}" ] ; then
    if ! mkdir "${TMPDIR2}" ; then
      echo "func_mk_tmpdirs(): mkdir ${TMPDIR2} failed"
      exit 1
    fi
  fi
}

function func_dir_from_pakname() {
  sed -e 's/[.]\(pak\|PAK\|Pak\)$//g'
}

function func_rm_tmpdirs() {
  # FIXME: checks
  if [ -d "${TMPDIR1}" ] ; then
    rm "${TMPDIR1}"/* 2> /dev/null ; rmdir "${TMPDIR1}"
  fi
  if [ -d "${TMPDIR2}" ] ; then
    rm "${TMPDIR2}"/* 2> /dev/null ; rmdir "${TMPDIR2}"
  fi
}

function func_clean() {
  rm "${LOGFILE}"
  func_rm_tmpdirs
  func_rm_testpak
  func_rm_testdir
}

# ========================================
# ========================================

do_tests
exit 0

# END
